use serde::{Serialize, Deserialize};
use std::{
	cell::RefCell,
	path::{Path, PathBuf},
	fs::File,
	time::SystemTime,
	io::Result,
	iter::Iterator
};

use super::{
	Archiveable,
	messages::Message
};

//use colored::*;

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Board {
	pub title: String,
	pub desc: String,
	pub messages: RefCell<Vec<Message>>,
	pub creation: u64,
	pub id: u32,
	pub path: String
}

impl Board {
	pub fn open<P>(path: P) -> Result<Board> where P: AsRef<Path> {
		match File::open(&path) {
			Ok(a) => Ok(serde_json::from_reader(a).unwrap()),
			Err(e) => Err(e)
		}
	}

	pub fn search<'a, I>(
		mut iter: I,
		id: u32
	) -> Option<&'a Board>
	where
		I: Iterator<Item = &'a Board>
	{
		iter.find(|x| x.id == id)
	}

	pub fn add_message(&self, msg: Message) {
		self.messages.borrow_mut().push(msg);
	}
}

impl Archiveable<Self> for Board {
	fn get_root(&self) -> &Path {
		Path::new(&self.path)
	}
}

#[derive(Debug, Serialize, Deserialize)]
pub struct BoardFactory {
	pub next_id: u32,
	pub data_dir: String
}

impl BoardFactory {
	pub fn new<P>(path: P) -> BoardFactory where P: AsRef<Path> {
		#[allow(clippy::redundant_field_names)]
		BoardFactory {
			next_id: 0,
			data_dir: path.as_ref().to_str().unwrap().into()
		}
	}

	pub fn assign_id(&mut self) -> u32 {
		let id = self.next_id;
		self.next_id += 1;
		id
	}
	pub fn set_data_dir<P>(&mut self, path: P) where P: AsRef<Path> {
		self.data_dir = path.as_ref().to_str().unwrap().into();
	}
	pub fn create_board<S>(&mut self, name: S, descript: S) -> Result<Board> where S: AsRef<str> {
		let mut path: PathBuf = PathBuf::from(self.data_dir.clone());
		path.push(self.next_id.to_string());
		#[allow(clippy::redundant_field_names)]
		let b = Board {
			title: String::from(name.as_ref()),
			desc: String::from(descript.as_ref()),
			messages: RefCell::new(Vec::new()),
			creation: SystemTime::now().duration_since(SystemTime::UNIX_EPOCH).unwrap().as_secs(),
			id: self.assign_id(),
			path: path.to_str().unwrap().to_string()
		};
		Ok(b)
		/*
		println!("board path: {:?}", path);
		fs::create_dir_all(&path).unwrap();
		path.push("board.json");
		Board::open(&path).or_else(|_e| -> Result<Board> {
			if !path.exists() {
				let mut file = File::create(&path).unwrap();
				file.write_all(serde_json::to_string(&b).unwrap().as_bytes())?;
			}
			Board::open(path.parent().unwrap())
		})
		*/
	}
}

impl Archiveable<Self> for BoardFactory {
	fn get_root(&self) -> &Path {
		Path::new(&self.data_dir)
	}
}
