use std::{
	env, fs, time,
	sync::Mutex,
	fs::{
		ReadDir
	},
	path::Path,
	time::{Duration, SystemTime}
};

use actix_web::{
	get, post, web,
	App, HttpResponse, HttpServer, Responder,
	cookie::Cookie
};

mod lib;
use lib::{
	*,
	Archiveable,
	boards::{
		Board,
		BoardFactory
	},
	users::UserManager,
	messages::MessageFactory
};

mod api;
use api::types::{AppState, CryptoUserForm};

/*
struct PageBuilder {
	body: &'static str,
	style: Option<String>,
	table_labels: Option<String>
}

impl PageBuilder {
	fn new(text: &'static str) -> PageBuilder {
		PageBuilder {
			body: text,
			style: None,
			table_labels: None
		}
	}

	fn build<S>(self, style: S, table_labels: S) -> String where S: AsRef<str> {
		String::from(self.body)
			.replacen("{}", style.as_ref(), 1)
			.replacen("{}", table_labels.as_ref(), 1)
	}
}
*/

#[post("/users/auth")]
pub async fn auth_html(
	data: web::Data<AppState>,
	form: web::Form<CryptoUserForm>
) -> HttpResponse {
	let user = data.user_manager.lock().unwrap().get_user_by_uname(form.username.clone());
	if let Some(u) = user {
		match u.verify(form.password.clone()) {
			Ok(a) => if a {
				let mut response = HttpResponse::Ok()
					.body(format!(r#"
<html><head>
<link rel="stylesheet" type="text/css" href="https://orbitalfox.us/Music/index.css">
<style>{}</style>
</head><body>
<div id="title" class="bar">
<h1>Hello, {1}</h1>
</div><div id="index">
<p>
Logged in as {1}, return to <a href="/boards/">index</a>
</p>
</body></html></div></body></html>
"#, include_str!("html/index.css"), u.username));
				response.add_cookie(
					&Cookie::build("auth", u.secret)
						//.domain("boards.solow.xyz")
						.domain("localhost")
						.path("*")
						//.same_site(actix_web::cookie::SameSite::Strict)
						//.secure(true)
						.finish()
				).unwrap();
				response
			} else {
				HttpResponse::Forbidden().body("Forbidden: Could not authenticate user")
			}
			Err(e) => HttpResponse::BadRequest().body(format!("Bad Request: {}", e))
		}
	} else {
		HttpResponse::BadRequest().body("Bad Request: User does not exist")
	}
}

#[get("/debug/")]
async fn debug(
	data: web::Data<AppState>
) -> impl Responder {
	HttpResponse::Ok().body(serde_json::to_string_pretty(data.as_ref()).unwrap())
}

#[get("/login/")]
async fn login() -> impl Responder {
	HttpResponse::Ok().body(format!(r#"
<html><head>
<link rel="stylesheet" type="text/css" href="https://orbitalfox.us/Music/index.css">
<style>{}</style>
</head><body>
<div id="title" class="bar">
<h1>Login</h1>
</div><div id="index">
<div class="auth">
<form action="/users/auth" method="post">
<div>
<label for="username">Username: </label>
<input type="text" id="username" name="username" required>
</div>
<div>
<label for="password">Password: </label>
<input type="password" id="password" name="password" required>
</div>
<input type="submit" value="Login">
</form>
</div>
</body></html></div></body></html>
"#, include_str!("html/index.css")
	))
}

#[get("/boards/")]
async fn list_boards(
	data: web::Data<AppState>,
) -> impl Responder {
	let now = time::SystemTime::now().duration_since(time::UNIX_EPOCH).unwrap();
	let mut ab = true;
	let boards = (*data.boards.lock().unwrap()).clone();
	HttpResponse::Ok().body(format!(
		r#"
<html><head>
<link rel="stylesheet" type="text/css" href="https://orbitalfox.us/Music/index.css">
<style>{}</style>
</head><body>
<div id="title" class="bar">
<h1>Board Index</h1>
</div><div id="index">
<table style="width: 100%;">
<tr><th>Name</th><th>Description</th><th>Last Active</th></tr>
{}
</table></body></html></div></body></html>
"#, include_str!("html/index.css"),
		boards.iter().map(|b| {
			ab = !ab;
			format!(
				"<tr class=\"{}\"><td><a href=\"/boards/{}/\">{}</a></td><td>{}</td><td>{}</td></tr>",
				if ab { "a" } else { "b" }, b.id, b.title, b.desc,
				if let Some(a) = b.messages.borrow().last() {
					format!("{:.1} minutes", now.checked_sub(Duration::from_secs(a.timestamp)).unwrap().as_secs()/60)
				} else {
					String::from("Never")
				}
			)
		}).collect::<String>()
	))
}

#[get("/boards/{board_id}/")]
async fn get_board(
	data: web::Data<AppState>,
	web::Path(board_id): web::Path<u32>
) -> HttpResponse {
	let mut ab = true;
	let boards = data.boards.lock().unwrap();
	if let Some(a) = Board::search(boards.iter(), board_id) {
		HttpResponse::Ok().body(format!(
			r#"
<html><head>
<link rel="stylesheet" type="text/css" href="https://orbitalfox.us/Music/index.css">
<style>
{}
</style>
</head>
<body>
<div id="title" class="bar">
<h1>
{}
</h1>
</div>
<div id="index">
<table style="width: 100%;">
{}
</table></body></html>
</div>
</body>
</html>
"#,
			include_str!("html/index.css"),
			a.title,
			a.messages.borrow().iter().map(|x| {
				ab = !ab;
				format!(
					"<tr class=\"message {}\"><td>{}</td><td>{}</td><td>{}</td></tr>",
					if ab { "a" } else { "b" },
					{
						let author = x.author.clone().into_inner();
						author.nick.clone().or_else(|| Some(author.username.clone())).unwrap()
					},
					x.text,
					x.timestamp
				)
			}).collect::<String>()
		))
	} else {
		HttpResponse::NotFound().body("That board could not be found")
	}
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
	println!("Initializing...");
	let data_dir = Path::new("data/");
	let mut pwd = env::current_dir().unwrap();

	let board_files: ReadDir = match fs::read_dir({
		pwd.push(data_dir);
		pwd.push("boards/");
		&pwd
	}) {
		Ok(a) => a,
		Err(_e) => {
			fs::create_dir_all(&pwd).unwrap();
			fs::read_dir(pwd).unwrap()
		}
	};

	let mut boards: Vec<Board> = Vec::new();

	for file in board_files.map(|x| x.unwrap()) {
		let mut path = file.path().to_path_buf();
		//let meta = fs::metadata(&path).unwrap();
		let ftype = file.file_type().unwrap();
		if ftype.is_dir() {
			path.push("board.json");
			println!("loading board: {:?}", path);
			boards.push(
				Board::open(&path).unwrap_or_else(|e| {
					panic!("Could not open board {:?}\n{}", path, e)
				})
			);
		}
	}

	let mut board_data = std::path::PathBuf::new();
	board_data.push(&data_dir);
	let ff = FactoryFactory::new(data_dir);
	let (mut bf, mf, um) = ff.build();
	board_data.push("boards/");
	bf.set_data_dir(board_data);

	let web_data = web::Data::new(AppState {
		boards: Mutex::new(boards.clone()),
		board_factory: Mutex::new(BoardFactory::load_state({
			println!("Loading board factory file...");
			let mut dd = data_dir.to_path_buf();
			dd.push("boards/factory.json");
			dd
		}).unwrap_or(bf)),
		msg_factory: Mutex::new(MessageFactory::load_state({
			println!("Loading message factory file...");
			let mut dd = data_dir.to_path_buf();
			dd.push("message_factory.json");
			dd
		}).unwrap_or(mf)),
		user_manager: Mutex::new(UserManager::load_state({
			println!("Loading user manager file...");
			let mut dd = data_dir.to_path_buf();
			dd.push("users.json");
			dd
		}).unwrap_or(um))
	});
	let wb = web_data.clone();

	println!("Finished initialization");
	println!("Starting server...");

	let server = HttpServer::new(move || {
		App::new()
			.app_data(web_data.clone())
			.service(debug)
			.service(list_boards)
			.service(get_board)
			.service(login)
			.service(auth_html)
			.service(
				web::scope("/api")
					.service(api::boards::new)
					.service(api::boards::list)
					.service(api::messages::new)
					.service(api::users::new)
					.service(api::users::list)
					.service(api::users::get)
					.service(api::users::auth)
			)
	}).bind("127.0.0.1:8080")?
		.run();
	println!("Started server");
	println!("{}", include_str!("res/banner"));
	let res = server.await;
	println!("\nExiting...");
	/*
	 * Note to self: handle all errors as much as possible
	 * Use unwrap() as little as possible
	 * If anything goes wrong while saving data data WILL get lost
	 */
	println!("Saving application state...");
	for b in wb.boards.lock().unwrap().iter() {
		fs::create_dir_all(&b.path).unwrap_or_else(
			|x| { println!("Could not create directory {:?}: {}", b.path, x) }
		);
		b.save_state("board.json").unwrap_or_else(
			|x| { println!("Could not save board {:?}: {}", b.title, x) }
		);
	}

	wb.board_factory
		.lock()
		.unwrap()
		.save_state("factory.json")
		.unwrap_or_else(
			|x| { println!("Could not save user factory state: {}", x) }
		);

	wb.msg_factory
		.lock()
		.unwrap()
		.save_state("message_factory.json")
		.unwrap_or_else(
			|x| { println!("Could not save message factory state: {}", x) }
		);

	wb.user_manager
		.lock()
		.unwrap()
		.save_state("users.json")
		.unwrap_or_else(
			|x| { println!("Could not save user data: {}", x) }
		);
	println!("Done!");
	res
}

// https://actix.rs/img/logo-large.png
